import logging
import os
import shutil
from typing import List, Tuple, Union

import cv2
import numpy as np

from xrprimer.data_structure.camera import FisheyeCameraParameter
from xrprimer.transform.camera.distortion import undistort_images
from xrprimer.utils.path_utils import (
    Existence,
    check_path_existence,
    prepare_output_path,
)
from .mview_pinhole_calibrator import MviewPinholeCalibrator
from .sview_fisheye_calibrator import SviewFisheyeDistortionCalibrator


class MviewFisheyeCalibrator(MviewPinholeCalibrator):

    def __init__(self,
                 chessboard_width: int,
                 chessboard_height: int,
                 chessboard_square_size: int,
                 work_dir: str,
                 calibrate_intrinsic: bool = False,
                 calibrate_distortion: bool = False,
                 calibrate_extrinsic: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Mview extrinsic calibrator for distorted fisheye cameras.

        Args:
            chessboard_width (int):
                How many internal corners along the
                horizontal edge of the chessboard.
            chessboard_height (int):
                How many internal corners along the
                vertical edge of the chessboard.
            chessboard_square_size (int):
                The edge length of a unit square in millimeter.
            work_dir (str):
                A path to an empty dir or nonexistent dir, undistorted
                frames will be saved there.
            calibrate_intrinsic (bool, optional):
                Whether to calibrate intrinsic. Defaults to False.
            calibrate_distortion (bool, optional):
                Whether to calibrate distortion coefficients.
                Defaults to False.
            calibrate_extrinsic (bool, optional):
                Whether to calibrate extrinsic. Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        MviewPinholeCalibrator.__init__(
            self,
            chessboard_width=chessboard_width,
            chessboard_height=chessboard_height,
            chessboard_square_size=chessboard_square_size,
            logger=logger)
        self.work_dir = work_dir
        if calibrate_intrinsic:
            self.logger.error('Intrinsic calibration not implemented yet.')
            raise NotImplementedError
        if (calibrate_intrinsic or calibrate_distortion
                or calibrate_extrinsic) is False:
            self.logger.error(
                'Please select at least one parameter to calibrate.')
            raise ValueError
        self.calibrate_intrinsic = calibrate_intrinsic
        self.calibrate_distortion = calibrate_distortion
        self.calibrate_extrinsic = calibrate_extrinsic

    def calibrate(
        self,
        frames: List[List[str]],
        fisheye_param_list: List[FisheyeCameraParameter],
    ) -> List[FisheyeCameraParameter]:
        """Calibrate multi-FisheyeCameraParameters with a chessboard. It takes
        intrinsics and distortion coefficients from fisheye_param_list,
        calibrates only extrinsics on undistorted frames.

        Args:
            frames (List[List[str]]):
                A nested list of distorted image paths. The shape is
                [n_frame, n_view], and each element is the path to
                an image file. '' stands for an empty image.
            fisheye_param_list (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameters. Intrinsic matrix
                and distortion coefficients are necessary for calibration.

        Returns:
            List[FisheyeCameraParameter]:
                A list of calibrated fisheye cameras, name, logger,
                resolution will be kept.
        """
        prepare_output_path(
            output_path=self.work_dir,
            path_type='dir',
            overwrite=False,
            logger=self.logger)
        if len(frames) <= 0:
            self.logger.error('Frames are necessary for fisheye extrinsic' +
                              ' calibration.')
            raise ValueError
        if len(frames[0]) != len(fisheye_param_list):
            self.logger.error(
                'n_view of frames must be equal to len(fisheye_param_list).')
            raise ValueError
        ret_list = [fisheye_param for fisheye_param in fisheye_param_list]
        if self.calibrate_distortion:
            sview_distortion_calibrator = SviewFisheyeDistortionCalibrator(
                chessboard_width=self.chessboard_width,
                chessboard_height=self.chessboard_height,
                logger=self.logger)
            for view_idx, fisheye_param in enumerate(ret_list):
                sview_frames = []
                for mview_frames in frames:
                    if len(mview_frames[view_idx]) > 0:
                        sview_frames.append(mview_frames[view_idx])
                calibrated_fisheye_param = \
                    sview_distortion_calibrator.calibrate(
                        frames=sview_frames,
                        fisheye_param=fisheye_param)
                ret_list[view_idx] = calibrated_fisheye_param
        if self.calibrate_extrinsic:
            undistorted_frames, pinhole_param_list = \
                self.__prepare_undistorted_frames__(
                    frames=frames,
                    fisheye_param_list=ret_list)
            pinhole_param_list = MviewPinholeCalibrator.calibrate(
                self,
                frames=undistorted_frames,
                pinhole_param_list=pinhole_param_list,
            )
            self.__remove_work_dir__()
            for view_idx, pinhole_param in enumerate(pinhole_param_list):
                fisheye_param = ret_list[view_idx].clone()
                fisheye_param.convention = 'opencv'
                fisheye_param.set_KRT(
                    R=pinhole_param.get_extrinsic_r(),
                    T=pinhole_param.get_extrinsic_t(),
                    world2cam=False)
                ret_list[view_idx] = fisheye_param
        return ret_list

    def __prepare_undistorted_frames__(
        self,
        frames: List[List[str]],
        fisheye_param_list: List[FisheyeCameraParameter],
    ) -> Tuple[List[List[str]], List[FisheyeCameraParameter]]:
        """Undistort images in frames, return paths to undistorted images and
        undistorted pinhole cameras.

        Args:
            frames (List[List[str]]):
                A nested list of distorted image paths. The shape is
                [n_frame, n_view], and each element is the path to
                an image file. '' stands for an empty image.
            fisheye_param_list (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameters. Intrinsic matrix
                and distortion coefficients are necessary for calibration.

        Returns:
            undistorted_frames (List[List[str]]):
                A nested list of undistorted image paths.
            pinhole_param_list (List[FisheyeCameraParameter]):
                A list of PinholeCameraParameters
        """
        n_view = len(fisheye_param_list)
        pinhole_param_list = []
        undistorted_frames = [[] for _ in range(len(frames))]
        for view_idx in range(n_view):
            sview_path_list = []
            default_shape = None
            for frame_idx, mview_paths in enumerate(frames):
                img_path = mview_paths[view_idx]
                if default_shape is None and len(img_path) > 0:
                    default_shape = cv2.imread(img_path).shape
                sview_path_list.append(img_path)
            sview_img_list = []
            sview_idx_list = []
            for frame_idx, img_path in enumerate(sview_path_list):
                if len(img_path) > 0:
                    img = cv2.imread(img_path)
                    sview_img_list.append(img)
                    sview_idx_list.append(frame_idx)
            sview_img_arr = np.asarray(sview_img_list)
            pinhole_param, undist_img_arr = undistort_images(
                distorted_cam=fisheye_param_list[view_idx],
                image_array=sview_img_arr)
            pinhole_param_list.append(pinhole_param)
            for img, frame_idx in zip(sview_img_list, sview_idx_list):
                path = os.path.join(
                    self.work_dir,
                    f'frame_{frame_idx:06d}_view_{view_idx:03d}.png')
                cv2.imwrite(path, img)
            for frame_idx, img_path in enumerate(sview_path_list):
                path = os.path.join(
                    self.work_dir,
                    f'frame_{frame_idx:06d}_view_{view_idx:03d}.png')
                if len(img_path) > 0:
                    undistorted_frames[frame_idx].append(path)
                else:
                    undistorted_frames[frame_idx].append('')
        return undistorted_frames, pinhole_param_list

    def __remove_work_dir__(self):
        """Remove temp files in self.work_dir."""
        if check_path_existence(self.work_dir) == \
                Existence.DirectoryExistEmpty or \
                check_path_existence(self.work_dir) == \
                Existence.DirectoryExistNotEmpty:
            shutil.rmtree(self.work_dir)



import logging
from typing import Union

from xrprimer.data_structure.camera import PinholeCameraParameter
from xrprimer.utils.log_utils import get_logger


class BaseCalibrator():

    def __init__(self,
                 work_dir: str = './temp',
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Base class of camera calibrator.

        Args:
            work_dir (str, optional):
                Path to the working dir for this instance.
                Defaults to './temp'.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        self.logger = get_logger(logger)
        self.work_dir = work_dir

    def calibrate(self) -> PinholeCameraParameter:
        """Calibrate a camera or several cameras. Input args shall not be
        modified and the calibrated camera will be returned.

        Returns:
            PinholeCameraParameter:
                The calibrated camera.
        """
        raise NotImplementedError



import os
import shutil

import pytest

from xrprimer.utils.path_utils import (
    Existence,
    check_path_existence,
    check_path_suffix,
)

input_dir = 'test/data/utils/test_path_utils'
output_dir = 'test/data/output/utils/test_path_utils'


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)


def test_check_path_existence():
    # test FileExist
    path = os.path.join(output_dir, 'FileExist.txt')
    with open(path, 'w') as f_write:
        f_write.write('\n')
    assert check_path_existence(path) == Existence.FileExist
    # test DirectoryExistEmpty
    path = os.path.join(output_dir, 'DirectoryExistEmpty')
    os.mkdir(path)
    assert check_path_existence(path) == Existence.DirectoryExistEmpty
    # test DirectoryExistNotEmpty
    path = os.path.join(output_dir, 'DirectoryExistNotEmpty')
    os.mkdir(path)
    with open(os.path.join(path, 'test_file.txt'), 'w') as f_write:
        f_write.write('\n')
    assert check_path_existence(path) == Existence.DirectoryExistNotEmpty
    # test MissingParent
    path = os.path.join(output_dir, 'MissingParent', 'test_file.txt')
    assert check_path_existence(path) == Existence.MissingParent
    # test DirectoryNotExist
    path = os.path.join(output_dir, 'DirectoryNotExist')
    assert check_path_existence(path) == Existence.DirectoryNotExist
    # test DirectoryNotExist
    path = os.path.join(output_dir, 'FileNotExist.txt')
    assert check_path_existence(path) == Existence.FileNotExist


def test_check_path_suffix():
    # test True
    path = os.path.join(output_dir, 'check_path_suffix.txt')
    with open(path, 'w') as f_write:
        f_write.write('\n')
    assert check_path_suffix(path, [])
    # test single str
    assert check_path_suffix(path, '.txt')
    # test list
    assert check_path_suffix(path, ['.txt'])
    assert check_path_suffix(path, ['.bin', '.txt'])
    # test upper case
    assert check_path_suffix(path, ['.bin', '.TXT'])
    # test dir
    path = os.path.join(output_dir, 'check_path_suffix_dir')
    os.mkdir(path)
    assert check_path_suffix(path, '')
    assert check_path_suffix(path, [''])
    assert not check_path_suffix(path, ['.txt'])



import logging
from typing import List, Union

import cv2
import numpy as np

from xrprimer.data_structure.camera import FisheyeCameraParameter
from .base_calibrator import BaseCalibrator


class SviewFisheyeDistortionCalibrator(BaseCalibrator):

    def __init__(self,
                 chessboard_width: int,
                 chessboard_height: int,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Single-view distortion calibrator for distorted fisheye camera. It
        takes an init intrinsic, fix it and calibrate distortion coefficients.

        Args:
            chessboard_width (int):
                How many internal corners along the
                horizontal edge of the chessboard.
            chessboard_height (int):
                How many internal corners along the
                vertical edge of the chessboard.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        BaseCalibrator.__init__(self, logger=logger)
        self.chessboard_width = chessboard_width
        self.chessboard_height = chessboard_height

    def calibrate(
        self,
        frames: List[str],
        fisheye_param: FisheyeCameraParameter,
    ) -> FisheyeCameraParameter:
        """Calibrate FisheyeCameraParameter with a chessboard. It takes
        intrinsics from fisheye_param, calibrates only distortion coefficients
        on undistorted frames.

        Args:
            frames (List[str]):
                A list of distorted image paths.
            fisheye_param (FisheyeCameraParameter):
                An instance of FisheyeCameraParameter. Intrinsic matrix
                is necessary for calibration, and the input instance
                will not be modified.

        Returns:
            FisheyeCameraParameter:
                An instance of FisheyeCameraParameter. Distortion coefficients
                are the only difference from input.
        """
        if len(frames) <= 0:
            self.logger.error('Frames are necessary for fisheye distortion' +
                              ' calibration.')
            raise ValueError
        # termination criteria
        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30,
                    0.001)
        # prepare object points in (x, y, 0),
        # like (0,0,0), (1,0,0), (2,0,0), ...
        sframe_obj_points = np.zeros(
            (self.chessboard_width * self.chessboard_height, 3), np.float32)
        sframe_obj_points[:, :2] = np.mgrid[
            0:self.chessboard_width,
            0:self.chessboard_height].T.reshape(-1, 2)
        mframe_obj_points = []  # 3d point in real world space
        mframe_img_points = []  # 2d points in image plane.
        for _, frame_path in enumerate(frames):
            img = cv2.imread(frame_path)
            if len(img.shape) == 3 and img.shape[-1] == 3:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            height, width = img.shape
            # Find the chess board corners
            pattern_found, corners = cv2.findChessboardCorners(
                img, (self.chessboard_width, self.chessboard_height), None)
            # If found, add object points, image points (after refining them)
            if pattern_found:
                mframe_obj_points.append(sframe_obj_points.copy())
                better_conners = cv2.cornerSubPix(img, corners, (11, 11),
                                                  (-1, -1), criteria)
                mframe_img_points.append(better_conners)
        # use preset-intrinsic and fix fx, fy, cx, cy
        calib_flags = cv2.CALIB_USE_INTRINSIC_GUESS + \
            cv2.CALIB_FIX_PRINCIPAL_POINT + \
            cv2.CALIB_FIX_FOCAL_LENGTH
        _, _, dist_coeff, _, _ = cv2.calibrateCamera(
            objectPoints=mframe_obj_points,
            imagePoints=mframe_img_points,
            imageSize=(width, height),
            cameraMatrix=np.asarray(fisheye_param.get_intrinsic(3)),
            distCoeffs=None,
            flags=calib_flags)
        ret_fisheye_param = fisheye_param.clone()
        dist_coeff = dist_coeff.tolist()[0]
        ret_fisheye_param.set_dist_coeff(
            dist_coeff_k=dist_coeff[:2] + dist_coeff[-1:],
            dist_coeff_p=dist_coeff[2:4])
        return ret_fisheye_param



import logging
import os
import shutil

import pytest

from xrprimer.utils.log_utils import get_logger, setup_logger

input_dir = 'test/data/utils/test_log_utils'
output_dir = 'test/data/output/utils/test_log_utils'


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)


def test_setup_logger():
    # test name
    test_logger = setup_logger(logger_name='test_name')
    assert get_logger('test_name') == test_logger
    # test level
    test_logger = setup_logger(
        logger_name='test_level', logger_level=logging.DEBUG)
    assert test_logger.level == logging.DEBUG
    # test level
    test_logger = setup_logger(
        logger_name='test_path',
        logger_path=os.path.join(output_dir, 'log.txt'))
    test_logger.info('test msg')
    assert os.path.exists(os.path.join(output_dir, 'log.txt'))
    # test format
    test_logger = setup_logger(
        logger_name='test_path',
        logger_format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    test_logger.info('test msg')


def test_get_logger():
    # test None
    test_logger = get_logger()
    assert test_logger.name == 'root'
    # test name
    test_logger = get_logger('any')
    assert test_logger.name == 'any'
    # test logger type
    test_logger = get_logger(test_logger)
    assert test_logger.name == 'any'



from .fisheye_camera import FisheyeCameraParameter
from .omni_camera import OmniCameraParameter
from .pinhole_camera import PinholeCameraParameter

__all__ = [
    'FisheyeCameraParameter', 'OmniCameraParameter', 'PinholeCameraParameter'
]



from xrprimer.data_structure.camera.camera import BaseCameraParameter
from .from_opencv import convert_camera_from_opencv
from .intrinsic import downgrade_k_4x4, upgrade_k_3x3
from .to_opencv import convert_camera_to_opencv

__all__ = ['upgrade_k_3x3', 'downgrade_k_4x4', 'convert_camera_parameter']


def convert_camera_parameter(
    cam_param: BaseCameraParameter,
    dst: str,
) -> BaseCameraParameter:
    """Convert a camera parameter instance into opencv convention.

    Args:
        cam_param (BaseCameraParameter):
            The input camera parameter, which is an instance of
            BaseCameraParameter subclass.
        dst (str):
            The name of destination convention.

    Returns:
        BaseCameraParameter:
            A camera in the same type as input, whose
            direction is same as cam_param, and convention
            equals to dst.
    """
    if cam_param.convention == dst:
        return cam_param
    else:
        opencv_cam = convert_camera_to_opencv(cam_param)
        dst_cam = convert_camera_from_opencv(opencv_cam, dst)
        if dst_cam.world2cam != cam_param.world2cam:
            dst_cam.inverse_extrinsic()
        return dst_cam
